import base64
import inspect
import logging
from types import ModuleType
from typing import Union


logger = logging.getLogger(__name__)


def make_udf(module_or_code: Union[ModuleType, str], *args) -> str:
    '''
    This function takes a Python module or a code string,
    encodes the source code of the module or the code string
    into a byte string, which is suitable for transmission over the Internet,
    and returns a simple Python code snipplet that decodes and executes
    the original source code.

    Args:
        `module_or_code`: either a Python module *object* (not the module name),
            or a code block as a string.

            If the UDF is module `mypackage.udfs.udf`, then you may do

                import mypackage.udfs.udf
                s = make_udf(mypackage.udfs.udf)

            but not

                s = make_udf('mypackage.udfs.udf')

            If the UDF is a script that can not be imported by the package
            mechanism, you can read the content of the script into a string
            and pass that to `make_udf`. For example, suppose the UDF is
            a script named 'myudf.py' in the same directory as the
            (script or module) code in which you call `make_udf`, you
            can do

                from pathlib import Path
                code = (Path(__file__).parent / 'myudf.py').read_text()
                s = make_udf(code)

        `args`: if present, positional arguments to the UDF script.
            Each argument value should be a simple string or number
            (which will be converted to string)
            that does not contain any special characters (such as space,
            quotes) that could potentially cause trouble on the command-line.

            In this case, `module_or_code` processes a known number 
            of arguments specified on the command-line before processing 
            records coming from `stdin`.
            The number of args must be fixed; there can not be optional 
            arguments with default values, because the UDF assumes data 
            records begin after the known number of arguments.

            See `hive_udf_args_example.py` for an example.

    Suppose `s` is the output of this function, then it is used like this
    to construct a HiveQL statement:

        sql = f"""
            SELECT
                TRANSFORM ( ...input_columns... )
                USING '{s}'
                AS (...output_columns...)
            FROM {db_name}.{table_name}
            WHERE ...
        """

    (The part after `FROM` may be a subquery rather than a table directly.)

    An important detail is that this string (`s`) is wrapped by single quotes
    in the HiveQL statement, as shown above.

    If `module_or_code` is a UDAF rather than UDF, a `CLUSTER BY` clause 
    is needed, like this:

        sql = f"""
            SELECT
                TRANSFORM (
                    ...input_columns_including_aggregation_columns...
                    )
                USING '{s}'
                AS (...output_columns...)
            FROM (
                SELECT
                    ...columns_including_aggregation_columns...
                FROM {db_name}.{table_name}
                CLUSTER BY ...aggregation_columns...
                ) AS t
        """

    This function supports 4 scenarios:

        1. UDF
        2. UDAF
        3. UDF with arguments
        4. UDAF with arguments

    The difference between UDF and UDAF does not appear in this function, 
    but rather in `module_or_code` and the HiveQL statements that use the 
    output of this function (i.e. presence of `CLUSTER BY`).

    While the current function is written in Python 3.6+, the UDF `module_or_code`
    is written in the Python version that is installed on the Hive server.
    '''

    if inspect.ismodule(module_or_code):
        s = inspect.getsource(module_or_code)
    else:
        assert isinstance(module_or_code, str)
        s = module_or_code

    encoded = base64.urlsafe_b64encode(s.encode())

    script = 'import sys, base64; code = sys.argv[1]; code = base64.urlsafe_b64decode(code); code = code if sys.version_info[0] == 2 else code.decode(); exec(code);'

    code = f'python -c "{script}" {str(encoded)[2:-1]}'

    if args:
        code = code + ' ' + ' '.join(str(v) for v in args)

    return code



'''
This is an example UDF for Hive, to be processed by `make_udf` and used in tests.

UDFs should follow this example. The main points include:

1. The code in this module is written against the Python version that is 
   available on the Hive cluster.

2. Because this module runs in the environment on the server (worker nodes 
   to be precise), only packages in the server's environment can be used.
   To be safe, it's a good idea to make this module *self sufficient* and use
   Python's standard libraries only. It should not use disk files.

3. This example defines a main function that processes a single line, 
   and the entry block passes individual lines on stdin to this function. 
   This is a reasonable pattern to use.

   However, if you have an existing UDF script that works but does not 
   follow this pattern, it is not necessary to modify the script. 
   It is expected that `make_udf` 'just works'.

4. `from __future__ import print_function` is recommended. 
   If this module is written against Python 2, but this module needs 
   to be imported into Python 3 code, then this import is required.

What this UDF does:

Take two columns: `id`, `info_json`.
Parse `info_json` to get `make` and `price`.
Print out these two columns.
If `price` does not exist in `make`, that column is NULL.
'''

from __future__ import print_function
import json
import sys

SEP = '\t'
NULL = '\\N'


def main():
    for line in sys.stdin:
        _, info_json = line.strip().split(SEP)
        info = json.loads(info_json)
        print(info['make'] + SEP + str(info.get('price', NULL)))


if __name__ == '__main__':
    main()



# Use should provide a conformant object `hive`
# and call the function `test_all` in their Hive setup
# to test the UDF functionalities.

import random
import string
from typing import List, Sequence, Tuple

import numpy as np
import pandas as pd
from hive_udf import (
    make_udf,
    hive_udf_example,
    hive_udaf_example,
    hive_udf_args_example)


class _Hive:
    def execute(self, sql: str):
        raise NotImplementedError

    def fetchall(self) -> List[tuple]:
        raise NotImplementedError

    def create_table(self, db_name, tb_name, cols: Sequence[Tuple[str, str]]):
        raise NotImplementedError


def _test_udf(hive, db_name, tb_name):
    print('records in table:')
    hive.execute(f'SELECT * FROM {db_name}.{tb_name} LIMIT 10')
    print(hive.fetchall())

    code = make_udf(hive_udf_example)
    sql = f'''
        SELECT
            TRANSFORM (
                id,
                info_json
            )
            USING '{code}'
            AS (make STRING, price FLOAT)
        FROM {db_name}.{tb_name}
        '''
    hive.execute(sql)
    rows = hive.fetchall()
    z = pd.DataFrame.from_records(
        list(rows), columns=['make', 'price'])
    z = z.sort_values(['make', 'price'])
    print('z')
    print(z)
    print('')

    # Expected result:
    #
    #     make   price
    # 1   ford  2000.0
    # 5   ford  4000.0
    # 2   ford     NaN
    # 0  honda  1000.0
    # 4  honda  2000.0
    # 3  tesla  3000.0

    assert len(z) == 6
    assert z['make'].tolist() == [
        'ford', 'ford', 'ford', 'honda', 'honda', 'tesla']
    assert np.isnan(z['price'].iloc[2])
    assert z['price'].iloc[3] == 1000


def _test_udaf(hive, db_name, tb_name):
    code = make_udf(hive_udaf_example)
    sql = f'''
        SELECT
            TRANSFORM (
                info_json
            )
            USING '{code}'
            AS (make STRING, avg_price FLOAT, null_prices INT)
        FROM (
            SELECT
                id,
                info_json
            FROM {db_name}.{tb_name}
            CLUSTER BY GET_JSON_OBJECT(info_json, '$.make')
            ) AS t
    '''
    hive.execute(sql)
    rows = hive.fetchall()
    z = pd.DataFrame.from_records(
        list(rows), columns=['make', 'avg_price', 'null_prices'])
    print('z')
    print(z)
    print('')

    # Expected result:
    #
    #     make  avg_price  null_prices
    # 1   ford     3000.0            1
    # 0  honda     1500.0            0
    # 2  tesla     3000.0            0

    assert len(z) == 3
    z = z.sort_values(['make'])
    assert z['make'].tolist() == ['ford', 'honda', 'tesla']
    assert z['avg_price'].tolist() == [3000, 1500, 3000]
    assert z['null_prices'].tolist() == [1, 0, 0]


def _run_sql(hive, sql, cols):
    hive.execute(sql)
    rows = hive.fetchall()
    z = pd.DataFrame.from_records(list(rows), columns=cols)
    z = z.sort_values(cols)
    print('z')
    print(z)
    print('')
    return z


def _test_udf_args(hive, db_name, tb_name):
    def make_sql(country, default_price):
        code = make_udf(hive_udf_args_example, country, default_price)
        sql = f'''
            SELECT
                TRANSFORM (
                    id,
                    info_json
                )
                USING '{code}'
                AS (make STRING, price FLOAT)
            FROM {db_name}.{tb_name}
            '''
        return sql

    ### jap ###

    sql = make_sql('jap', 250)
    z = _run_sql(hive, sql, ['make', 'price'])

    # Expected result:
    #
    #     make   price
    # 0  honda  1000.0
    # 1  honda  2000.0

    assert z['make'].tolist() == ['honda', 'honda']
    assert z['price'].tolist() == [1000, 2000]

    ### america ###

    sql = make_sql('america', 550)
    z = _run_sql(hive, sql, ['make', 'price'])

    # Expected result:
    #
    #     make   price
    # 1   ford   550.0
    # 0   ford  2000.0
    # 3   ford  4000.0
    # 2  tesla  3000.0

    assert z['make'].tolist() == ['ford', 'ford', 'ford', 'tesla']
    assert z['price'].tolist() == [550, 2000, 4000, 3000]

    sql = make_sql('all', 340)
    z = _run_sql(hive, sql, ['make', 'price'])

    # Expected result:
    #
    #    make   price
    # 2  ford   340.0
    # 1  ford  2000.0
    # 5  ford  4000.0
    # 0 honda  1000.0
    # 4 honda  2000.0
    # 3 tesla  3000.0

    assert z['make'].tolist() == ['ford', 'ford', 'ford',
                                  'honda', 'honda', 'tesla']
    assert z['price'].tolist() == [340, 2000, 4000, 1000, 2000, 3000]


def _test_udf_followed_by_agg(hive, db_name, tb_name):
    code = make_udf(hive_udf_example)
    sql = f'''
        SELECT
            make,
            SUM(price) AS price_total
        FROM (
            SELECT
                TRANSFORM (
                    id,
                    info_json
                )
                USING '{code}'
                AS (make STRING, price FLOAT)
            FROM {db_name}.{tb_name}
            ) A
        GROUP BY make
        '''
    z = _run_sql(hive, sql, ['make', 'price_total'])

    # Expected result:
    #
    #     make price_total
    # 0   ford      6000.0
    # 1  honda      3000.0
    # 2  tesla      3000.0

    assert len(z) == 3
    assert z['make'].tolist() == ['ford', 'honda', 'tesla']
    assert z['price_total'].tolist() == [6000.0, 3000.0, 3000.0]


def _make_tmp_name():
    return 'tmp_' + ''.join(random.choices(string.ascii_lowercase, k=20))


def _get_databases(hive: _Hive):
    hive.execute('SHOW DATABASES')
    z = hive.fetchall()
    return [v[0] for v in z]


def _get_tables(hive: _Hive, db_name: str):
    hive.execute(f'SHOW TABLES IN {db_name}')
    z = hive.fetchall()
    return [v[0] for v in z]


def test_all(hive: _Hive, db_name: str):
    assert db_name in _get_databases(hive)

    tb_name = _make_tmp_name()

    print('creating table "{}.{}"'.format(db_name, tb_name))
    hive.execute(f'DROP TABLE IF EXISTS {db_name}.{tb_name}')

    # hive.execute(f'''
    #              CREATE TABLE {db_name}.{tb_name}
    #              (
    #                  id INT,
    #                  info_json STRING
    #              )
    #              STORED AS ORC
    #              TBLPROPERTIES (
    #                  'orc.compress'='ZLIB'
    #              )
    #              ''')
    hive.create_table(
        db_name,
        tb_name,
        [('id', 'INT'), ('info_json', 'STRING')],
    )

    try:
        assert tb_name in _get_tables(hive, db_name)
        print(f'table {db_name}.{tb_name} created successfully')

        hive.execute(f'''
                     INSERT OVERWRITE TABLE {db_name}.{tb_name}
                     VALUES
                        (1, '{{"make": "honda", "price": 1000}}'),
                        (2, '{{"make": "ford", "price": 2000}}'),
                        (3, '{{"make": "ford"}}'),
                        (4, '{{"make": "tesla", "price": 3000}}'),
                        (5, '{{"make": "honda", "price": 2000}}'),
                        (6, '{{"make": "ford", "price": 4000}}')
                     ''')

        _test_udf(hive, db_name, tb_name)
        _test_udaf(hive, db_name, tb_name)
        _test_udf_args(hive, db_name, tb_name)
        _test_udf_followed_by_agg(hive, db_name, tb_name)
    finally:
        print(f'dropping table {db_name}.{tb_name}')
        hive.execute(f'DROP TABLE IF EXISTS {db_name}.{tb_name}')
        assert tb_name not in _get_tables(hive, db_name)



'''
This is an example UDAF for Hive, to be processed by `make_udf` and used in tests.

What this UDAF does:

Take a single column `info_json`, which is `clustered`
by the `make` element in the JSON (i.e. records of the same `make`
come consecutively to the same Hive worker node),
parse this JSON to get `make` and `price`.
For each `make`, calculate average price of the records that have `price`;
also count the records that lack `price`.
Print out three columns:

    `make`, `average_price`, `number_of_null_price`

Since this UDAF requires the records to be already clustered by `make`,
the caller necessarily needs to extract `make` from `info_json` and use it.
So, the caller may as well provide two columns: `make` and `info_json`.
That would simplify this UDAF a little bit.

Here we assume the caller does not provide that convenience.
'''

from __future__ import print_function
import itertools
import json
import sys


SEP = '\t'
NULL = '\\N'


def read_data(input_data=sys.stdin):
    for line in input_data:
        info = json.loads(line.strip())
        yield info['make'], info.get('price', NULL)


def main():
    data = read_data()
    for make, group in itertools.groupby(data, lambda x: x[0]):
        n_good = 0
        n_bad = 0
        price_sum = 0
        for _, price in group:
            if price == NULL:
                n_bad += 1
            else:
                p = float(price)
                # In reality one may want to handle possible error here.
                # Here we assume this will not fail.

                price_sum += p
                n_good += 1
        avg_price = round(price_sum / n_good)
        print(make + SEP + str(avg_price) + SEP + str(n_bad))


if __name__ == '__main__':
    main()



import pathlib
from setuptools import setup, find_packages


HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

version = '1.1.3'

setup(
    name='hive-udf',
    version=version,
    description='Hive UDF (user defined functions)',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/zpz/hive-udf',
    license='MIT',
    python_requires='>=3.6',
    package_dir={'': 'src'},
    packages=['hive_udf'],
    include_package_data=True,
)



'''
This is an example UDAF for Hive, to be processed by `make_udf` and used in tests.

Reference:
    https://florianwilhelm.info/2016/10/python_udf_in_hive/


What this UDAF does:


Take two columns: `id`, `info_json`.
Parse `info_json` to get `make` and `price`.
Print out these two columns.

Two parameters control further details:

The first parameter is `country`, with possible values 'all', 'jap', 'america'.
Only records with car maker in the specified country are processed.

The second parameter is 'default price'. If `info_json` does not contain price,
use this default value.
'''

from __future__ import print_function
import json
import sys


SEP = '\t'
NULL = '\\N'


MAKES = {
    'jap': ['honda'],
    'america': ['ford', 'tesla']
}


def main(country, price_default):
    for line in sys.stdin:
        _, info_json = line.strip().split(SEP)
        info = json.loads(info_json)
        make = info['make']
        price = info.get('price', price_default)
        if country == 'all' or make in MAKES[country]:
            print(make + SEP + str(price))


if __name__ == '__main__':
    # Note: due to the way `make_udf` handles this script,
    # the parameters should be retrieved as `sys.argv[2]`
    # and `sys.argv[3]`, etc.

    country = sys.argv[2]
    price_default = float(sys.argv[3])

    # In reality this error situation should be handled in a better way.
    assert country == 'all' or country in MAKES

    main(country, price_default)
